/*
 * 
 * @author Manuel Parras  
 */
package proyecto.consola;

import java.awt.Component;
import java.awt.Font;
import java.util.ArrayList;

import javax.swing.JTextPane;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.plaf.ComponentUI;
import javax.swing.text.BadLocationException;
import javax.swing.text.html.HTMLEditorKit;
import org.apache.log4j.Logger;

import proyecto.editor.Editor;
import proyecto.utilidades.HojaEstilo;
import proyecto.utilidades.MenuPopupConsola;


/**
 * <p>
 * ConsolaTextPane
 * </p>
 * <p>
 * Esta clase es la contenedora del todo texto referente a la compilacin de un
 * elemento, siempre que sea de la extensin .java o .jj.
 * </p>
 * <p>
 * Si se produjera un error en el momento de la compilacin esta clase nos crea
 * una enlace con el fichero y la posicin del error. Estas manipulaciones la
 * manufactura nuestra clase interna llamada LinkHtmlListener. La clase
 * LinkHtmlListener lo que interpreta es nuestro objeto CompilacionDatos para
 * obtener la ubicacin del problema y fichero
 * </p>
 * <p>Si este componente se cerrara en la aplicacin, perderamos todas las
 * referencias que se hayan realizado. </p>
 * 
 * @see proyecto.consola.CompilacionDatos
 *  
 * @see proyecto.consola.ConsolaTextPane.LinkHtmlListener
 */

public class ConsolaTextPane extends JTextPane {

	/** Constante serialVersionUID. */
	private static final long serialVersionUID = 1L;

	/** Constante fuente que contedra nuestro texto FUENTE_COMPILACION. */
	private static final Font FUENTE_COMPILACION = new Font("Arial", 0, 12);

	/** objLogger. */
	private static final Logger objLogger = Logger
			.getLogger(ConsolaTextPane.class);

	/** @see proyecto.consola.CompilacionDatos */
	private ArrayList <CompilacionDatos> icompilacionDatos;

	/**
	 * <p>
	 * Constructor de clase.
	 * </p>
	 * Inicializa el editorkit valido para el reconocimiento de html.
	 * 
	 * <p>
	 * Creado : Parras, 24-jun-2009
	 * </p>
	 * 
	 * @see javax.swing.text.html.HTMLEditorKit
	 */
	public ConsolaTextPane() {

		HTMLEditorKit kit = new HTMLEditorKit();
		kit.setStyleSheet(HojaEstilo.get());
		this.setEditorKit(kit);
		this.setContentType("text/html");
		this.setEditable(false);
		this.icompilacionDatos = new ArrayList<CompilacionDatos>();
		this.addMouseListener(new MenuPopupConsola(this));

	}

	/**
	 * Evitamos que nuestra consola realize LineWrap.
	 * 
	 * @return the scrollable tracks viewport width
	 */
	public boolean getScrollableTracksViewportWidth() {

		Component parent = this.getParent();
		ComponentUI ui = this.getUI();

		return (ui.getPreferredSize(this).width <= parent.getSize().width);
	}

	/**
	 * Activar hyper link listener.
	 * 
	 * @param aCompilacionDatos
	 *            the a compilacion datos
	 * 
	 *            <p>
	 *            Creado : Parras, 24-jun-2009
	 *            </p>
	 */
	public void activarHyperLinkListener(ArrayList<CompilacionDatos> aCompilacionDatos) {
		this.addHyperlinkListener(new LinkHtmlListener(aCompilacionDatos));
	}

	/**
	 * Copiar un componente ConsolaTextPane pasadado como parametro y activa su
	 * listener si tiene datos de compilacin.
	 * 
	 * @param aConsola
	 *            ConsolaTextPane
	 */
	public void copiarConsolaTextPane(ConsolaTextPane aConsola) {

		aConsola.setCompilacionDatos(this.getCompilacionDatos());
		aConsola.setContentType("text/html");
		try {
			aConsola.setText("<html><body><DIV STYLE=\"font-family:"
					+ FUENTE_COMPILACION.getFontName()
					+ "; font-size:"
					+ FUENTE_COMPILACION.getSize()
					+ "px; color: black\">"
					+ this.getDocument().getText(0,
							this.getDocument().getLength())
					+ "</div></body></html>");
			if (objLogger.isDebugEnabled()) {
				objLogger.debug("\nTexto Consola :  "
						+ "<html><body><DIV STYLE=\"font-family:"
						+ FUENTE_COMPILACION.getFontName()
						+ "; font-size:"
						+ FUENTE_COMPILACION.getSize()
						+ "px; color: black\">"
						+ this.getDocument().getText(0,
								this.getDocument().getLength())
						+ "</div></body></html>");
			}
		} catch (BadLocationException e) {
			if (objLogger.isDebugEnabled()) 
				objLogger.debug(" CopiarConsolaTextPane "+e);
		}
			
		if (aConsola.getCompilacionDatos() != null ){
			//for (CompilacionDatos compilacion : aConsola.getCompilacionDatos()){
			
				//if ( compilacion.esValido() )
						aConsola.activarHyperLinkListener(aConsola.getCompilacionDatos());
			//}
			
		}

	}

	/**
	 * Habilitar a nuesrto consola con la propiedad de Copiar.
	 *<p>
	 * Creado : Parras, 24-jun-2009
	 * </p>
	 */
	public void copiar() {

		this.copy();

	}

	/**
	 * Habilitar a nuesrto consola con la propiedad de Pegar.
	 *<p>
	 * Creado : Parras, 24-jun-2009
	 * </p>
	 */
	public void pegar() {
		this.paste();
	}

	/**
	 * Habilitar a nuesrto consola con la propiedad de Cortar.
	 * <p>
	 * Creado : Parras, 24-jun-2009
	 * </p>
	 */
	public void cortar() {
		this.cut();
	}

	/**
	 * Habilitar a nuesrto consola con la propiedad de Selecionar todo.
	 * <p>
	 * Creado : Parras, 24-jun-2009
	 * </p>
	 */
	public void selecionarTodo() {
		this.selectAll();
	}

	/**
	 * Habilitar a nuesrto consola con la propiedad de Limpiar texto.
	 * <p>
	 * Creado : Parras, 24-jun-2009
	 * </p>
	 */
	public void limpiarTexto() {
		this.setText("");
	}

	/**
	 * The listener interface for receiving linkHtml events. The class that is
	 * interested in processing a linkHtml event implements this interface, and
	 * the object created with that class is registered with a component using
	 * the component's <code>addLinkHtmlListener<code> method. When
	 * the linkHtml event occurs, that object's appropriate
	 * method is invoked.
	 * 
	 * @see javax.swing.event.HyperlinkEvent
	 */
	class LinkHtmlListener implements HyperlinkListener {

		/** @see proyecto.consola.CompilacionDatos. */
		ArrayList<CompilacionDatos> iCompilacionDatos;

		/**
		 * Constructor link html listener.
		 * 
		 * Creado : Parras, 24-jun-2009
		 * 
		 * @param aCompilacionDatos
		 *           CompilacionDatos
		 */
		public LinkHtmlListener(ArrayList<CompilacionDatos> aCompilacionDatos) {
			iCompilacionDatos = aCompilacionDatos;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see
		 * javax.swing.event.HyperlinkListener#hyperlinkUpdate(javax.swing.event
		 * .HyperlinkEvent)
		 */
		public void hyperlinkUpdate(HyperlinkEvent evt) {
			if (evt.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
				
				String lValor = evt.getDescription();
				if (lValor != null)
					lValor = lValor.trim();
				
				if (lValor != null &&  lValor.matches("[0-9]+") && iCompilacionDatos.size() >= new Integer(lValor).intValue()
						 && iCompilacionDatos.get(new Integer(lValor).intValue()).esValido() )
					Editor.get().getPanel().moverCursorAError(iCompilacionDatos.get(new Integer(lValor).intValue()));
			}
			
				
		}

	}

	/**
	 * Obtener el compilacion datos.
	 * 
	 * @return el compilacion datos Creado : Parras, 24-jun-2009
	 */
	public ArrayList <CompilacionDatos> getCompilacionDatos() {
		return icompilacionDatos;
	}

	/**
	 * Modificar la propiedad compilacion datos.
	 * 
	 * @param acompilacionDatos
	 *            CompilacionDatos
	 */
	public void setCompilacionDatos( ArrayList <CompilacionDatos> acompilacionDatos) {
		// si existen datos de compilacion debemos activar su litserner para
		// poder interpretarlos
		
		this.icompilacionDatos = acompilacionDatos;
	}
	
	
	/**
	 * Aade compilacion.
	 * 
	 * @param acompilacionDatos
	 *            the acompilacion datos
	 * 
	 * @return  int en la posicin que se ha insertado 
	 * Creado : Parras, 14-feb-2010
	 */
	public int addCompilacion(CompilacionDatos acompilacionDatos) {
		// si existen datos de compilacion debemos activar su litserner para
		// poder interpretarlos
		acompilacionDatos.setPosicion(icompilacionDatos.size());
		this.icompilacionDatos.add(acompilacionDatos);
		return acompilacionDatos.getPosicion();
	}

}