/*
 * 
 * @author Manuel Parras  
 */
package proyecto.editor;

import java.awt.Component;
import java.awt.FontMetrics;
import java.awt.Insets;
import java.io.IOException;
import java.io.Reader;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.SortedSet;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.util.Vector;

import javax.swing.JTextPane;
import javax.swing.plaf.ComponentUI;
import javax.swing.text.AbstractDocument;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultStyledDocument;
import javax.swing.text.Element;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;
import javax.swing.text.TabSet;
import javax.swing.text.TabStop;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;

import org.apache.log4j.Logger;

import proyecto.sintaxis.JavaCCLexer;
import proyecto.undo.CompoundUndoManager;
import proyecto.utilidades.MenuPopup;
import proyecto.utilidades.OperacionesTextPane;
import proyecto.utilidades.UtilidadesEpsilon;

import com.Ostermiller.Syntax.Lexer.Lexer;
import com.Ostermiller.Syntax.Lexer.Token;

/**
 * <p>
 * El componente de texto que hemos seleccionado admite texto con cierto
 * formato: texto plano, HTML y RTF por defecto, aunque en nuestro caso lo hemos
 * ampliado para nuestro uso especfico. Tambin nos admite mezclar fuentes y
 * colores, dichas caractersticas asociadas a las cadenas de texto se realizan
 * a travs de los estilos. Un estilo es un conjunto de caractersticas como
 * son: FontSize, ForegroundColor, isBold, isItalic, etc. y se manipula a travs
 * de constantes del estilo( StyleConstants.setForeground( estilo,Color.red )).
 * </p>
 * <p>
 * Para todas estas necesidades ampliamos la clase DefaultStyledDocument para
 * poder colorear las palabras segn nos indique nuestro analizador lxico.
 * </p>
 * <p>
 * El coloreado lo hacemos a travs del Threard Colorer que es el encargado de
 * llamar a lxico para que nos defina el estilo de pintado. La llamada al
 * coloreado lo hacemos desde la subclase HighLightedDocument en los mtodos de
 * insertString
 * <p>
 * <p>
 * De todo el proceso del reconocedor se encarga la librera agregada a nuestro
 * proyecto syntax.jar.
 * <p>
 * <p>
 * Vamos a comentar algunas caractersticas importantes de la librera:
 * </p>
 * <ul>
 * <li>El analizador lxico devuelve un token y contina analizando con el
 * siguiente token encontrado.</li>
 * <li>El analizador lxico puede ser reiniciado desde cualquier punto sin
 * afectar al color de las palabras que nos encontramos posteriormente.</li>
 * <li>Podemos determinar el ltimo token generado y el estado que se encuentra
 * el que estamos analizando</li>
 * <li>Cuando modificamos un palabra en medio del documento necesitamos hacer un
 * seguimiento de todos los estados inciales susceptibles a cambios, de manera
 * que podamos reiniciar el lxico en cualquier parte del documento.</li>
 * <li>Si el analizador lxico vuelve al mismo estado que estuvo la ltima vez
 * significa que el token est coloreado correctamente.</li>
 * </ul>
 * 
 * @see javax.swing.JTextPane
 * @author Parras
 */
public class JavaTextPane extends JTextPane {

	/** Constant serialVersionUID. */
	private static final long serialVersionUID = 1L;
	// Estilo de Contexto
	// private static StyleContext styleContext = new StyleContext();
	// private UndoManager undoManager;
	private static final Logger objLogger = Logger.getLogger(JavaTextPane.class);
	/** undo manager. */
	private CompoundUndoManager undoManager;

	/** ib cambios. */
	private boolean ibCambios;

	/** ip panel edicion. */
	PanelEdicion ipPanelEdicion;
	// private StyledEditorKit editorKit;

	// static
	// {
	// initStyles();
	// }
	// Documento Con estilos
	/** document. */
	protected HighLightedDocument document;
	// Documento Para leer para Lexer
	/** document reader. */
	protected DocumentReader documentReader;
	// Eligue el color que necesitamos
	/** @see com.Ostermiller.Syntax.Lexer.Lexer */
	protected Lexer syntaxLexer;
	// Thread Para colorear
	/** colorer. */
	protected Colorer colorer;
	// Objeto Para sincronizar las clases
	/** doclock. */
	private Object doclock = new Object();

	/** texto form impri. */
	protected StringBuffer textoFormImpri;

	/** ok imprimir. */
	private boolean okImprimir = false;

	/** root element. */
	private Element rootElement;

	/**
	 * Constructor java text pane.
	 * 
	 * Creado : Parras, 07-nov-2009
	 */
	public JavaTextPane() {
		super();
	}

	/**
	 * Constructor java text pane.
	 * 
	 * Creado : Parras, 07-nov-2009
	 * 
	 * @param aPanel
	 *            a panel
	 * @param aStyleContext
	 *            a style context
	 * @param texto
	 *            texto
	 */
	public JavaTextPane(PanelEdicion aPanel, StyleContext aStyleContext,
			String texto) {
		super();

		document = new HighLightedDocument(aStyleContext);
		setDocument(document);
		setTabs(this, 4);
		setCaretPosition(0);
		setMargin(new Insets(5, 5, 5, 5));
		colorer = new Colorer();
		colorer.start();
		// create new document.
		documentReader = new DocumentReader(this.document);

		syntaxLexer = new JavaCCLexer(documentReader);
		
		ibCambios = true;

		// this.getDocument().addUndoableEditListener(getUndoManager());
		this.addMouseListener(new MenuPopup(this, aPanel));

		ipPanelEdicion = aPanel;
		rootElement = document.getDefaultRootElement();
		this.setText(texto);
		setUndoManager();
	}

	/**
	 * Modifica la propiedad undo manager.
	 * 
	 * Creado : Parras, 07-nov-2009
	 */
	public void setUndoManager() {
		if (undoManager == null)
			undoManager = new CompoundUndoManager(this);
	}

	/**
	 * setTabs
	 * 
	 * <p>
	 * Establece a nuestro JTextPane los tabuladores, no cuenta por carecteres
	 * sino por pxeles.
	 * </p>
	 * 
	 * @param textPane
	 *            textPane
	 * @param aCaracteresParaTab
	 *            caracteres para los tab
	 */
	public void setTabs(JTextPane textPane, int aCaracteresParaTab) {
		FontMetrics fm = textPane.getFontMetrics(textPane.getFont());
		int charWidth = fm.charWidth('w');
		int tabWidth = charWidth * aCaracteresParaTab;

		TabStop[] tabs = new TabStop[10];

		for (int j = 0; j < tabs.length; j++) {
			int tab = j + 1;
			// System.out.print("\nTabuladores : "+tab * tabWidth);
			tabs[j] = new TabStop(tab * tabWidth);

		}

		
		TabSet tabSet = new TabSet(tabs);
		SimpleAttributeSet attributes = new SimpleAttributeSet();
		StyleConstants.setTabSet(attributes, tabSet);

		int length = textPane.getDocument().getLength();
		textPane.getStyledDocument().setParagraphAttributes(0, length,
				attributes, false);
	}

	// Para colocar el caret en la linea que queremos
	/**
	 * Modificar la propiedad del caret.
	 * <p>
	 * Para colocar el caret en la linea que queremos
	 * </p>
	 * 
	 * @param linenumber
	 *            Numero linea
	 */
	public void setCursorANumeroLinea(int linenumber) {
		setCaretPosition(0);
		if (linenumber < 2)
			return;
		StringTokenizer st = new StringTokenizer(getText(), "\n", true);
		int count = 0;
		while ((st.hasMoreTokens() & (linenumber > 1))) {
			String s = st.nextToken();
			count += s.length();
			if (s.equals("\n"))
				linenumber--;
		}
		setCaretPosition(count);
	}

	//	

	/**
	 * (non-Javadoc) Forma de hacer LineWrap al TextPane
	 * 
	 * @see javax.swing.JEditorPane#getScrollableTracksViewportWidth()
	 */
	public boolean getScrollableTracksViewportWidth() {

		Component parent = this.getParent();
		ComponentUI ui = this.getUI();

		return (ui.getPreferredSize(this).width <= parent.getSize().width);
	}

	// Forma de hacer LineWrap al TextPane

	/*
	 * public boolean getScrollableTracksViewportWidth() {
	 * 
	 * return (this.getSize().width < getParent().getSize().width);
	 * 
	 * }
	 * 
	 * public void setSize(Dimension d) {
	 * 
	 * if (d.width < getParent().getSize().width) { d.width =
	 * getParent().getSize().width;
	 * 
	 * } super.setSize(d); }
	 */

	/**
	 * Obtener el panel edicion.
	 * 
	 * @return el panel edicion Creado : Parras, 07-nov-2009
	 */
	public PanelEdicion getPanelEdicion() {
		return ipPanelEdicion;
	}

	/**
	 * Obtener el cambios.
	 * 
	 * @return el cambios Creado : Parras, 07-nov-2009
	 */
	public boolean getCambios() {
		return ibCambios;
	}

	/**
	 * Modificar la propiedad cambios.
	 * 
	 * @param aValor
	 *            La propiedad cambios
	 */
	public void setCambios(boolean aValor) {
		ibCambios = aValor;
	}

	/**
	 * Undo.
	 */
	public void deshacer() {
		/*
		 * if (getUndoManager().canUndo()) { getUndoManager().undo(); }
		 */
		try {
			undoManager.undo();
			this.requestFocus();
		} catch (CannotUndoException ex) {			
			if (objLogger.isDebugEnabled())
				objLogger.debug("Unable to undo: " + ex);
		}
	}

	/**
	 * formatearCodigo.
	 */
	public void formatearCodigo() {
		int liPosicionCursor = this.getCaretPosition();
		this.setText(UtilidadesEpsilon.formatearCodigo(this.getText(),
				getPanelEdicion().getFicheroTabRun()));
		OperacionesTextPane.irALinea(this, liPosicionCursor);
	}

	/**
	 * Redo.
	 */
	public void rehacer() {
		/*
		 * if (getUndoManager().canRedo()) { getUndoManager().redo(); }
		 */
		try {
			undoManager.redo();
			this.requestFocus();
		} catch (CannotRedoException ex) {
			if (objLogger.isDebugEnabled())
				objLogger.debug("Unable to redo: " + ex);
		}
	}

	/**
	 * Copiar. Creado : Parras, 07-nov-2009
	 */
	public void copiar() {

		this.copy();

	}

	/**
	 * Pegar. Creado : Parras, 07-nov-2009
	 */
	public void pegar() {
		this.paste();
	}

	/**
	 * Cortar. Creado : Parras, 07-nov-2009
	 */
	public void cortar() {
		this.cut();
	}

	/**
	 * Selecionar todo. Creado : Parras, 07-nov-2009
	 */
	public void selecionarTodo() {
		this.selectAll();
	}

	/**
	 * Suprimir. Creado : Parras, 07-nov-2009
	 */
	public void suprimir() {
		this.replaceSelection(null);
	}

	// Colorear y poner en un treeSet la palabra coloreada con las posiciones

	/**
	 * Class Colorer.
	 */
	private class Colorer extends Thread {

		// Guarda las Iniciales posiciones y las mete ordenadas. Para tener
		// referencias a ellas por
		// Si son modificadas
		/** ini positions. */
		private TreeSet iniPositions = new TreeSet(new DocPositionComparator());

		// Quita la posiciones invalidas y busca la posiciones validas
		/** new positions. */
		private HashSet newPositions = new HashSet();

		/**
		 * A simple wrapper representing something that needs to be colored.
		 * Placed into an object so that it can be stored in a Vector.
		 */
		// Simple envoltotio que necesitamos para colorear
		private class RecolorEvent {

			/** position. */
			public int position;

			/** adjustment. */
			public int adjustment;

			/**
			 * Constructor recolor event.
			 * 
			 * Creado : Parras, 07-nov-2009
			 * 
			 * @param position
			 *            position
			 * @param adjustment
			 *            adjustment
			 */
			public RecolorEvent(int position, int adjustment) {
				this.position = position;
				this.adjustment = adjustment;
			}
		}

		// Vector que nos guarda la comunicacion entre la theards
		/** Vector que nos guarda la comunicacion entre la theards */
		private volatile Vector v = new Vector();

		// Nos guarda la ultima posicion seleccionada
		/** Nos guarda la ultima posicion seleccionada */
		private volatile int change = 0;

		// La ultima posicion coloreada
		/** La ultima posicion coloreada */
		private volatile int lastPosition = -1;

		/** asleep. */
		private volatile boolean asleep = false;

		// Sincronizar los vectores
		/** Sincronizar los vectores. */
		private Object lock = new Object();

		// Selecciona la hebra y mira la posicion del documento. Proceso fifo
		/**
		 * Color.
		 * 
		 * <p>
		 * Selecciona la hebra y mira la posicion del documento. Proceso fifo
		 * </p>
		 * 
		 * 
		 * 
		 * @param position
		 *            position
		 * @param adjustment
		 *            adjustment Creado : Parras, 07-nov-2009
		 */
		public void color(int position, int adjustment) {
			// figure out if this adjustment effects current run.
			// if it does, then adjust place in document
			// that gets highlighted.
			if (position < lastPosition) {
				if (lastPosition < position - adjustment) {
					change -= lastPosition - position;
				} else {
					change += adjustment;
				}
			}
			synchronized (lock) {
				v.add(new RecolorEvent(position, adjustment));
				if (asleep) {
					this.interrupt();
				}
			}
		}

		/**
		 * colorer runs forever and may sleep for long periods of time. It
		 * should be interrupted every time there is something for it to do.
		 */
		public void run() {
			int position = -1;
			int adjustment = 0;
			// if we just finish, we can't go to sleep until we
			// ensure there is nothing else for us to do.
			// use try again to keep track of this.
			boolean tryAgain = false;
			for (;;) { // Bucle Infinito
				synchronized (lock) {
					if (v.size() > 0) {
						RecolorEvent re = (RecolorEvent) (v.elementAt(0));
						v.removeElementAt(0);
						position = re.position;
						adjustment = re.adjustment;
					} else {
						tryAgain = false;
						position = -1;
						adjustment = 0;
					}
				}
				if (position != -1) {
					SortedSet workingSet;
					Iterator workingIt;
					DocPosition startRequest = new DocPosition(position);
					DocPosition endRequest = new DocPosition(position
							+ ((adjustment >= 0) ? adjustment : -adjustment));
					DocPosition dp;
					DocPosition dpStart = null;
					DocPosition dpEnd = null;

					// find starting position. We must start at least one
					// token before current position
					try {
						// all good positions before
						workingSet = iniPositions.headSet(startRequest);
						// last of stuff before
						dpStart = ((DocPosition) workingSet.last());
					} catch (NoSuchElementException x) {
						// if there were no good positions before requested
						// start,
						// we can always start at very beginning.
						dpStart = new DocPosition(0);
					}

					// if stuff was removed, take any removed positions off
					// list.
					if (adjustment < 0) {
						workingSet = iniPositions.subSet(startRequest,
								endRequest);
						workingIt = workingSet.iterator();
						while (workingIt.hasNext()) {
							workingIt.next();
							workingIt.remove();
						}
					}

					// adjust positions of everything after insertion/removal.
					workingSet = iniPositions.tailSet(startRequest);
					workingIt = workingSet.iterator();
					while (workingIt.hasNext()) {
						((DocPosition) workingIt.next())
								.adjustPosition(adjustment);
					}

					// now go through and highlight as much as needed
					workingSet = iniPositions.tailSet(dpStart);
					workingIt = workingSet.iterator();
					dp = null;
					if (workingIt.hasNext()) {
						dp = (DocPosition) workingIt.next();
					}
					try {
						Token t;
						boolean done = false;
						dpEnd = dpStart;
						synchronized (doclock) {
							// we are playing some games with lexer for
							// efficiency.
							// we could just create a new lexer each time here,
							// but instead,
							// we will just reset it so that it thinks it is
							// starting at the
							// beginning of document but reporting a funny start
							// position.
							// Reseting lexer causes close() method on reader
							// to be called but because close() method has no
							// effect on the
							// DocumentReader, we can do this.
							syntaxLexer.reset(documentReader, 0, dpStart
									.getPosition(), 0);
							// After lexer has been set up, scroll reader so
							// that it
							// is in correct spot as well.
							documentReader.seek(dpStart.getPosition());
							// we will highlight tokens until we reach a good
							// stopping place.
							// first obvious stopping place is end of document.
							// lexer will return null at end of document and wee
							// need to stop there.
							t = syntaxLexer.getNextToken();
							//System.out.print(" \nToken " + t);
						}
						newPositions.add(dpStart);
						while (!done && t != null) {
							// this is actual command that colors stuff.
							// Color stuff with description of style matched
							// to hash table that has been set up ahead of time.
							synchronized (doclock) {
								if (t.getCharEnd() <= document.getLength()) {

									Style style = getStyle(t.getDescription());
									document.setCharacterAttributes(t
											.getCharBegin()
											+ change, t.getCharEnd()
											- t.getCharBegin(), style, true);									
									// record position of last bit of text that
									// we colored
									dpEnd = new DocPosition(t.getCharEnd());
								}
								lastPosition = (t.getCharEnd() + change);
							}
							// other more complicated reason for doing no more
							// highlighting
							// is that all colors are same from here on out
							// anyway.
							// We can detect this by seeing if place that lexer
							// returned
							// to initial state last time we highlighted is same
							// as the
							// place that returned to initial state this time.
							// As long as that place is after last changed text,
							// everything
							// from there on is fine already.
							if (t.getState() == Token.INITIAL_STATE) {
								// System.out.println(t);
								// look at all positions from last time that are
								// less than or
								// equal to current position
								while (dp != null
										&& dp.getPosition() <= t.getCharEnd()) {
									if (dp.getPosition() == t.getCharEnd()
											&& dp.getPosition() >= endRequest
													.getPosition()) {
										// we have found a state that is same
										done = true;
										dp = null;
									} else if (workingIt.hasNext()) {
										// didn't find it, try again.
										dp = (DocPosition) workingIt.next();
									} else {
										// didn't find it, and there is no more
										// info from last
										// time. This means that we will just
										// continue
										// until end of document.
										dp = null;
									}
								}
								// so that we can do this check next time,
								// record all the
								// initial states from this time.
								newPositions.add(dpEnd);
							}
							synchronized (doclock) {
								t = syntaxLexer.getNextToken();
							}
						}

						// remove all old initial positions from place where
						// we started doing highlighting right up through last
						// bit of text we touched.
						workingIt = iniPositions.subSet(dpStart, dpEnd)
								.iterator();
						while (workingIt.hasNext()) {
							workingIt.next();
							workingIt.remove();
						}

						// Remove all positions that are after end of file.:
						workingIt = iniPositions.tailSet(
								new DocPosition(document.getLength()))
								.iterator();
						while (workingIt.hasNext()) {
							workingIt.next();
							workingIt.remove();
						}

						// and put new initial positions that we have found on
						// list.
						iniPositions.addAll(newPositions);
						newPositions.clear();

						/*
						 * workingIt = iniPositions.iterator(); while
						 * (workingIt.hasNext()){
						 * System.out.println(workingIt.next()); }
						 * 
						 * System.out.println("Started: " +
						 * dpStart.getPosition() + " Ended: " +
						 * dpEnd.getPosition());
						 */
					} catch (IOException x) {
					}
					synchronized (doclock) {
						lastPosition = -1;
						change = 0;
					}
					// since we did something, we should check that there is
					// nothing else to do before going back to sleep.
					tryAgain = true;
				}
				asleep = true;
				if (!tryAgain) {
					try {
						sleep(0xffffff);
					} catch (InterruptedException x) {
					}

				}
				asleep = false;

			}

		}

	}

	/**
	 * COloreamos entero nuestro documento.
	 */
	public void colorAll() {

		color(0, document.getLength());
	}

	/**
	 * Color a section of document. actual coloring will start somewhere before
	 * requested position and continue as long as needed.
	 * 
	 * @param position
	 *            starting point for coloring.
	 * @param adjustment
	 *            amount of text inserted or removed at starting point.
	 */
	public void color(int position, int adjustment) {
		colorer.color(position, adjustment);

	}

	/**
	 * Just like a DefaultStyledDocument but intercepts inserts and removes to
	 * color them.
	 */
	private class HighLightedDocument extends DefaultStyledDocument {

		/**
		 * 
		 */
		private static final long serialVersionUID = 4475604318769628865L;

		/**
		 * Constructor high lighted document.
		 * 
		 * Creado : Parras, 07-nov-2009
		 * 
		 * @param styles
		 *            styles
		 */
		public HighLightedDocument(StyleContext styles) {
			super(styles);
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see javax.swing.text.AbstractDocument#insertString(int,
		 * java.lang.String, javax.swing.text.AttributeSet)
		 */
		public void insertString(int offs, String str, AttributeSet a)
				throws BadLocationException {
			// Aqu le damos habilidades inteligentes a nuestro editor
			if (str.equals("{")) // Despues de una { no la cierra
				str = addMatchingBrace(offs);
			else if (str.equals("\n"))// despues de \n no debe colocar el cursor
				// a la misma altura
				str = calculoEspacios(offs);

			super.insertString(offs, str, a);

			color(offs, str.length());
			documentReader.update(offs, str.length());
		}

		/**
		 * Calcula los espacios donde debemos seguir escribiendo.
		 * 
		 * @param offset
		 *            offset
		 * 
		 * @return string
		 * 
		 * @throws BadLocationException
		 *             bad location exception
		 */
		protected String calculoEspacios(int offset)
				throws BadLocationException {
			StringBuffer whiteSpace = new StringBuffer();
			int line = rootElement.getElementIndex(offset);
			int i = rootElement.getElement(line).getStartOffset();

			while (true) {
				String temp = document.getText(i, 1);

				if (temp.equals(" ") || temp.equals("\t")) {
					whiteSpace.append(temp);
					i++;
				} else
					break;
			}

			return "\n" + whiteSpace.toString();
		}

		/**
		 * Aade matching brace.
		 * 
		 * @param offset
		 *            offset
		 * 
		 * @return string Creado : Parras, 07-nov-2009
		 * 
		 * @throws BadLocationException
		 *             bad location exception
		 */
		protected String addMatchingBrace(int offset)
				throws BadLocationException {
			StringBuffer whiteSpace = new StringBuffer();
			int line = rootElement.getElementIndex(offset);
			int i = rootElement.getElement(line).getStartOffset();

			while (true) {
				String temp = document.getText(i, 1);

				if (temp.equals(" ") || temp.equals("\t")) {
					whiteSpace.append(temp);
					i++;
				} else
					break;
			}

			return "{\n" + whiteSpace.toString() + "\t\n"
					+ whiteSpace.toString() + "}";
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see javax.swing.text.AbstractDocument#remove(int, int)
		 */
		public void remove(int offs, int len) throws BadLocationException {
			super.remove(offs, len);
			color(offs, -len);
			documentReader.update(offs, -len);
		}
	}

	/**
	 * A wrapper for a position in a document appropriate for storing in a
	 * collection.
	 */
	class DocPosition {

		/** actual position. */
		private int position;

		/**
		 * Get position represented by this DocPosition.
		 * 
		 * @return position
		 */
		int getPosition() {
			return position;
		}

		/**
		 * Construct a DocPosition from given offset into document.
		 * 
		 * @param position
		 *            position this DocObject will represent
		 */
		public DocPosition(int position) {
			this.position = position;
		}

		/**
		 * Adjust this position. This is useful in cases that an amount of text
		 * is inserted or removed before this position.
		 * 
		 * @param adjustment
		 *            amount (either positive or negative) to adjust this
		 *            position.
		 * 
		 * @return DocPosition, adjusted properly.
		 */
		public DocPosition adjustPosition(int adjustment) {
			position += adjustment;
			return this;
		}

		/**
		 * Two DocPositions are equal iff they have same internal position.
		 * 
		 * @param obj
		 *            obj
		 * 
		 * @return if this DocPosition represents same position as another.
		 */
		public boolean equals(Object obj) {
			if (obj instanceof DocPosition) {
				DocPosition d = (DocPosition) (obj);
				if (this.position == d.position) {
					return true;
				} else {
					return false;
				}
			} else {
				return false;
			}
		}

		/**
		 * A string representation useful for debugging.
		 * 
		 * @return A string representing position.
		 */
		public String toString() {
			return "" + position;
		}
	}

	/**
	 * A comparator appropriate for use with Collections of DocPositions.
	 */
	class DocPositionComparator implements Comparator {

		/**
		 * Does this Comparator equal another? Since all DocPositionComparators
		 * are same, they are all equal.
		 * 
		 * @param obj
		 *            obj
		 * 
		 * @return true for DocPositionComparators, false otherwise.
		 */
		public boolean equals(Object obj) {
			if (obj instanceof DocPositionComparator) {
				return true;
			} else {
				return false;
			}
		}

		/**
		 * Compare two DocPositions.
		 * 
		 * @param o1
		 *            first DocPosition
		 * @param o2
		 *            second DocPosition
		 * 
		 * @return negative if first < second, 0 if equal, positive if first >
		 *         second
		 */
		public int compare(Object o1, Object o2) {
			if (o1 instanceof DocPosition && o2 instanceof DocPosition) {
				DocPosition d1 = (DocPosition) (o1);
				DocPosition d2 = (DocPosition) (o2);
				return (d1.getPosition() - d2.getPosition());
			} else if (o1 instanceof DocPosition) {
				return -1;
			} else if (o2 instanceof DocPosition) {
				return 1;
			} else if (o1.hashCode() < o2.hashCode()) {
				return -1;
			} else if (o2.hashCode() > o1.hashCode()) {
				return 1;
			} else {
				return 0;
			}
		}
	}

	/**
	 * A reader interface for an abstract document. Since syntax highlighting
	 * packages only accept Stings and Readers, this must be used. Since close()
	 * method does nothing and a seek() method has been added, this allows us to
	 * get some performance improvements through reuse. It can be used even
	 * after lexer explicitly closes it by seeking to place that we want to read
	 * next, and reseting lexer.
	 */
	class DocumentReader extends Reader {

		

		/**
		 * Current position in document. Incremented whenever a character is
		 * read.
		 */
		private long position = 0;

		/** Saved position used in mark and reset methods. */
		private long mark = -1;

		/** document that we are working with. */
		private AbstractDocument document;

		/**
		 * Construct a reader on given document.
		 * 
		 * @param document
		 *            document to be read.
		 */
		public DocumentReader(AbstractDocument document) {
			this.document = document;
		}

		
		/**
		 * Modifying document while reader is working is like pulling rug out
		 * from under reader. Alerting reader with this method (in a nice thread
		 * safe way, this should not be called at the same time as a read)
		 * allows reader to compensate.
		 * 
		 * @param position
		 *            position
		 * @param adjustment
		 *            adjustment
		 */
		public void update(int position, int adjustment) {
			if (position < this.position) {
				if (this.position < position - adjustment) {
					this.position = position;
				} else {
					this.position += adjustment;
				}
			}
		}
		
		/**
		 * Has no effect. This reader can be used even after it has been closed.
		 */
		public void close() {
		}

		/**
		 * Save a position for reset.
		 * 
		 * @param readAheadLimit
		 *            ignored.
		 */
		public void mark(int readAheadLimit) {
			mark = position;
		}

		/**
		 * This reader support mark and reset.
		 * 
		 * @return true
		 */
		public boolean markSupported() {
			return true;
		}

		/**
		 * Read a single character.
		 * 
		 * @return character or -1 if end of document has been reached.
		 */
		public int read() {
			if (position < document.getLength()) {
				try {
					char c = document.getText((int) position, 1).charAt(0);
					position++;
					return c;
				} catch (BadLocationException x) {
					return -1;
				}
			} else {
				return -1;
			}
		}

		/**
		 * Read and fill buffer. This method will always fill buffer unless end
		 * of document is reached.
		 * 
		 * @param cbuf
		 *            buffer to fill.
		 * 
		 * @return number of characters read or -1 if no more characters are
		 *         available in document.
		 */
		public int read(char[] cbuf) {
			return read(cbuf, 0, cbuf.length);
		}

		/**
		 * Read and fill buffer. This method will always fill buffer unless end
		 * of document is reached.
		 * 
		 * @param cbuf
		 *            buffer to fill.
		 * @param off
		 *            offset into buffer to begin fill.
		 * @param len
		 *            maximum number of characters to put in buffer.
		 * 
		 * @return number of characters read or -1 if no more characters are
		 *         available in document.
		 */
		public int read(char[] cbuf, int off, int len) {
			if (position < document.getLength()) {
				int length = len;
				if (position + length >= document.getLength()) {
					length = document.getLength() - (int) position;
				}
				if (off + length >= cbuf.length) {
					length = cbuf.length - off;
				}
				try {
					String s = document.getText((int) position, length);
					position += length;
					for (int i = 0; i < length; i++) {
						cbuf[off + i] = s.charAt(i);
					}
					return length;
				} catch (BadLocationException x) {
					return -1;
				}
			} else {
				return -1;
			}
		}

		/**
		 * Ready.
		 * 
		 * @return true
		 */
		public boolean ready() {
			return true;
		}

		/**
		 * Reset this reader to last mark, or beginning of document if a mark
		 * has not been set.
		 */
		public void reset() {
			if (mark == -1) {
				position = 0;
			} else {
				position = mark;
			}
			mark = -1;
		}

		/**
		 * Skip characters of input. This method will always skip maximum number
		 * of characters unless end of file is reached.
		 * 
		 * @param n
		 *            number of characters to skip.
		 * 
		 * @return actual number of characters skipped.
		 */
		public long skip(long n) {
			if (position + n <= document.getLength()) {
				position += n;
				return n;
			} else {
				long oldPos = position;
				position = document.getLength();
				return (document.getLength() - oldPos);
			}
		}

		/**
		 * Seek to given position in document.
		 * 
		 * @param n
		 *            offset to which to seek.
		 */
		public void seek(long n) {
			if (n <= document.getLength()) {
				position = n;
			} else {
				position = document.getLength();
			}
		}
	}

	/**
	 * Obtener el texto formateado imprir.
	 * 
	 * @return el texto formateado imprir Creado : Parras, 07-nov-2009
	 */
	public StringBuffer getTextoFormateadoImprir() {
		textoFormImpri = new StringBuffer();
		okImprimir = true;
		UtilidadesEpsilon.cursorEspera(ipPanelEdicion, true);

		imprimir();

		// okImprimir=false;
		UtilidadesEpsilon.cursorEspera(ipPanelEdicion, false);
		return textoFormImpri;
	}

	/**
	 * Imprimir. Creado : Parras, 07-nov-2009
	 */
	public void imprimir() {

		Token t;
		DocPosition dpStart = null;
		try {
			syntaxLexer.reset(documentReader, 0, 0, 0);
			// After lexer has been set up, scroll reader so that it
			// is in correct spot as well.
			documentReader.seek(0);
			t = syntaxLexer.getNextToken();

			boolean ok = false;

			while (t != null && !ok) {

				if (t.getCharEnd() <= document.getLength()) {

					Style style = getStyle(t.getDescription());
					document.setCharacterAttributes(t.getCharBegin(), t
							.getCharEnd()
							- t.getCharBegin(), style, true);
					// System.out.print("\nbollImprimir-->"+getOkImprimir());

					// Para el Imprimir
					// "<font color=\"#FF0000\">"+element+" </font>"
					try {
						// String
						// rgb=Integer.toHexString(StyleConstants.getForeground(style).getRed())
						// +Integer.toHexString(StyleConstants.getForeground(style).getGreen())
						// +Integer.toHexString(StyleConstants.getForeground(style).getBlue());

						String rgb = Integer.toHexString(StyleConstants
								.getForeground(style).getRGB());
						rgb = rgb.substring(2, rgb.length());
						String slTexto = document.getText(t.getCharBegin(), t
								.getCharEnd()
								- t.getCharBegin());
						// System.out.print(slTexto);
						// if (slTexto.equals("<"))
						// slTexto="&#60;";
						// if (slTexto.equals(">"))
						// slTexto="&#62;";
						// if (slTexto.equals("&"))
						// slTexto="&amp;";

						// Si tenemos un comentario tenemos que parsearlo a
						// parte,
						// porque nos dara confliptos con las palabras
						// reservadas del ireport.
						// if (t.getDescription().equals("comment")
						// ||t.getDescription().equals("comment")){
						slTexto = slTexto.replaceAll("&", "&amp;");
						slTexto = slTexto.replaceAll("<", "&#60;");
						slTexto = slTexto.replaceAll(">", "&#62;");
						//
						// }
						textoFormImpri.append("<font color=" + "\"" + "#" + rgb
								+ "\"" + ">" + slTexto + "</font>");
						// System.out.print("<font color="+"\""+"#"+rgb+"\""+">"+slTexto+"</font>");
					} catch (BadLocationException e) {
						e.printStackTrace();
					}
					t = syntaxLexer.getNextToken();
				} else {
					ok = true;
				}
			}
		} catch (IOException e1) {
			// TODO Bloque catch generado automticamente
			e1.printStackTrace();
		}

	}

	/**
	 * Obtener el ok imprimir.
	 * 
	 * @return el ok imprimir Creado : Parras, 07-nov-2009
	 */
	public boolean getOkImprimir() {
		return okImprimir;
	}

	/**
	 * Modificar la propiedad ok imprimir.
	 * 
	 * @param okImprimir
	 *            La propiedad ok imprimir
	 */
	public void setOkImprimir(boolean okImprimir) {
		this.okImprimir = okImprimir;
	}

}
